/* Public domain. */
/* Based on tumsoul.py by Maxime Bizon. */

#include "sgetopt.h"
#include "bytestr.h"
#include "allreadwrite.h"
#include "buffer.h"
#include "fmtscan.h"
#include "strerr2.h"
#include "gen_alloc.h"
#include "stralloc.h"
#include "djbunix.h"
#include "skamisc.h"
#include "md5.h"
#include "netbuffer.h"

char const *PROG = "soulspark" ;
#define USAGE "soulspark [ -c | -s ] [ -f pwdfile ] [ -t timeout ] user prog..."

#define USER_AGENT "soulspark-0.12"


static void expectrep (buffer_ref b)
{
  satmp.len = 0 ;
  if (sanitize_read(skagetln(b, &satmp, '\n')) <= 0)
    strerr_diefu1sys(111, "read a line from fd 6") ;
  satmp.s[--satmp.len] = 0 ;
  if (!str_start(satmp.s, "rep 002 --"))
    strerr_dief4x(111, "weird server answer: expected ", "rep 002 --", " got ", satmp.s) ;
}

int main (int argc, char const *const *argv, char const *const *envp)
{
  char const *pwdfile = 0 ;
  unsigned char flagserver = 0 ;
  unsigned int timeout = 20 ;
  {
    int opt ;
    while ((opt = subgetopt(argc, argv, "csf:t:")) != opteof)
      switch (opt)
      {
        case 'c' : flagserver = 0 ; break ;
        case 's' : flagserver = 1 ; break ;
        case 'f' : pwdfile = optarg ; break ;
        case 't' : if (!uint0_scan(optarg, &timeout)) strerr_dieusage(100, USAGE) ;
        default : strerr_dieusage(100, USAGE) ;
      }
    argc -= optind ;
    argv += optind ;
  }
  if (argc < 2) strerr_dieusage(100, USAGE) ;

  {
    char buf6[BUFFER_INSIZE_SMALL] ;
    char buf7[BUFFER_OUTSIZE_SMALL] ;
    char pwd[512] ;
    buffer b6 = BUFFER_INIT(&buffer_netread, 6, buf6, BUFFER_INSIZE_SMALL) ;
    buffer b7 = BUFFER_INIT((iofunc_t_ref)&buffer_netwrite, 7, buf7, BUFFER_OUTSIZE_SMALL) ;
    int pwdlen ;
    int fdpwd = 3 ;

    if (pwdfile)
    {
      fdpwd = open_readb(pwdfile) ;
      if (fdpwd == -1)
        strerr_diefu3sys(111, "open ", pwdfile, " for reading") ;
    }
    pwdlen = allread(fdpwd, pwd, 512) ;
    if (pwdlen < 0) strerr_diefu1sys(111, "read password") ;
    fd_close(fdpwd) ;
    netbuffer_settimeout(timeout) ;

    {
      MD5Schedule hash = MD5_INIT() ;
      unsigned int i, j ;
      satmp.len = 0 ;
      if (sanitize_read(skagetln(&b6, &satmp, '\n')) <= 0)
        strerr_diefu1sys(111, "read a line from fd 6") ;
      satmp.s[--satmp.len] = 0 ;
      i = byte_chr(satmp.s, satmp.len, ' ') + 1 ;
      if (i >= satmp.len) strerr_dief2x(111, "invalid server line: ", satmp.s) ;
      i += byte_chr(satmp.s + i, satmp.len - i, ' ') + 1 ;
      if (i >= satmp.len) strerr_dief2x(111, "invalid server line: ", satmp.s) ;
      j = byte_chr(satmp.s + i, satmp.len - i, ' ') ;
      md5_update(&hash, satmp.s + i, j) ;
      md5_update(&hash, "-", 1) ;
      i += j + 1 ;
      if (i >= satmp.len) strerr_dief2x(111, "invalid server line: ", satmp.s) ;
      j = byte_chr(satmp.s + i, satmp.len - i, ' ') ;
      md5_update(&hash, satmp.s + i, j) ;
      md5_update(&hash, "/", 1) ;
      i += j + 1 ;
      if (i >= satmp.len) strerr_dief2x(111, "invalid server line: ", satmp.s) ;
      j = byte_chr(satmp.s + i, satmp.len - i, ' ') ;
      md5_update(&hash, satmp.s + i, j) ;
      i += j + 1 ;
      if (i >= satmp.len) strerr_dief2x(111, "invalid server line: ", satmp.s) ;
      md5_update(&hash, pwd, pwdlen) ;
      md5_final(&hash, pwd + 32) ;
      ucharn_fmt(pwd, pwd + 32, 16) ;
    }

    if (buffer_putsflush(&b7, "auth_ag ext_user none none\n") == -1)
      strerr_diefu1sys(111, "send auth_ag to server") ;

    expectrep(&b6) ;

    if ((buffer_putsalign(&b7, "ext_user_log ") == -1)
     || (buffer_putsalign(&b7, argv[0]) == -1)
     || (buffer_putalign(&b7, " ", 1) == -1)
     || (buffer_putalign(&b7, pwd, 32) == -1)
     || (buffer_putalign(&b7, " ", 1) == -1)
     || (buffer_putsalign(&b7, flagserver ? "Server" : "Client") == -1)
     || (buffer_putsalign(&b7, "%20mode ") == -1)
     || (buffer_putsalign(&b7, USER_AGENT) == -1)
     || (buffer_putflush(&b7, "\n", 1) == -1))
      strerr_diefu1sys(111, "send ext_user_log to server") ;

    expectrep(&b6) ;

    if (buffer_putsflush(&b7, "user_cmd state server\n") == -1)
      strerr_diefu1sys(111, "send user_cmd to server") ;
  }
  satmp.len = 0 ;

  pathexec_run(argv[1], argv+1, envp) ;
  strerr_dieexec(111, argv[1]) ;
}
