/* Public domain. */

#include <errno.h>
#include "uint16.h"
#include "sgetopt.h"
#include "fmtscan.h"
#include "strerr2.h"
#include "gen_alloc.h"
#include "stralloc.h"
#include "uint16alloc.h"
#include "socket.h"
#include "djbunix.h"

char const *PROG = "tcpping" ;
#define USAGE "tcpping [ -x | -X ] [ -T x+y ] ip:port [ ip:port... ]"

#define DEFAULTPORT 22

static void usage ()
{
  strerr_dieusage(100, USAGE) ;
}

static void dieparse (unsigned int i)
{
  char fmt[FMT_ULONG] ;
  fmt[uint_fmt(fmt, i)] = 0 ;
  strerr_dief2x(100, "parse error at argument ", fmt) ;
}

static int doit (char const *ip, uint16 port, unsigned int timeout, unsigned char flagstrict)
{
  int r ;
  int s = socket_tcp() ;
  if (s == -1) strerr_diefu1sys(111, "create socket") ;
  r = socket_timeoutconn(s, ip, port, timeout) ;
  {
    register int e = errno ;
    fd_close(s) ;
    errno = e ;
  }
  switch (r)
  {
    case 1 : return 1 ;
    case -1 : strerr_diefu1sys(111, "attempt to connect") ;
    default : return !flagstrict && (errno == ECONNRESET) ;
  }
}

int main (int argc, char const *const *argv)
{
  stralloc ips = GEN_ALLOC_ZERO ;
  uint16alloc ports = GEN_ALLOC_ZERO ;
  unsigned int x = 2, y = 58 ;
  unsigned int i = 0 ;
  unsigned char flagmustexist = 1 ;
  {
    register int opt ;
    while ((opt = subgetopt(argc, argv, "xXT:")) != opteof)
      switch (opt)
      {
        case 'x' : flagmustexist = 1 ; break ;
        case 'X' : flagmustexist = 0 ; break ;
        case 'T' :
        {
          unsigned int n = uint_scan(optarg, &x) ;
          if (!n) usage() ;
          if (!optarg[n])
          {
            y = 0 ;
            break ;
          }
          if (optarg[n] != '+') usage() ;
          if (!uint0_scan(optarg + n + 1, &y)) usage() ;
          break ;
        }
        default : usage() ;
      }
    argc -= optind ;
    argv += optind ;
  }
  if (!argc) usage() ;

  if (!stralloc_ready(&ips, argc<<2)
   || !uint16alloc_ready(&ports, argc))
    strerr_diefu1sys(111, "parse command line") ;

  ports.len = argc ;
  ips.len = ports.len << 2 ;
  for (; i < ports.len ; i++)
  {
    unsigned int n = ip4_scan(argv[i], ips.s + (i << 2)) ;
    if (!n) dieparse(i) ;
    if (!argv[i][n]) ports.s[i] = DEFAULTPORT ;
    else if (argv[i][n] != ':') dieparse(i) ;
    else if (!uint160_scan(argv[i] + n + 1, ports.s + i)) dieparse(i) ;
  }

  for (i = 0 ; i < ports.len ; i++)
    if (doit(ips.s + (i << 2), ports.s[i], x, flagmustexist)) return 0 ;
  if (y)
    for (i = 0 ; i < ports.len ; i++)
      if (doit(ips.s + (i << 2), ports.s[i], y, flagmustexist)) return 0 ;
  return 1 ;
}
